# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_utils.ipynb.

# %% auto 0
__all__ = ['ut_interp', 'ut_interp_mw', 'ut_interp_rho', 'u_fwhm_interp', 'ut_interp_nfw', 'ut_interp_nfw_101', 'ut_interp_axion',
           'ut_interp_dressed', 'm_low_interp', 'm_high_interp', 'rho_min', 'rho_max', 'rhos', 'A_thresh_min',
           'A_thresh_max', 'A_threshs', 'dist_mw', 'dist_m31', 'density_mw', 'density_m31', 'mass_enclosed_mw',
           'mass_enclosed_m31', 'velocity_dispersion_mw', 'velocity_dispersion_m31', 'dist', 'einstein_rad',
           'velocity_radial', 'event_duration', 'get_primed_coords', 'scientific_format', 'w_func', 'rho_func',
           'magnification', 'magnification_wave', 'displacement', 'integrand_polar_wave', 'integrand_polar',
           'magnification_finite_wave', 'magnification_finite', 'u_t_finite', 'u_t_point', 'u_t_finite_wave',
           'make_ut_interp', 'u_t_nfw', 'u_t_nfw_101', 'u_t_axion', 'u_t_axion_101', 'u_t_dressed', 'u_t_dressed_101',
           'u_t_nfw_thresh', 'u_t_axion_thresh', 'u_t_dressed_thresh', 'ut_func_new', 'f_func', 't_fwhm_gondolo',
           'uhm_point', 't_e_from_fwhm']

# %% ../nbs/04_utils.ipynb 3
from .parameters import *
import numpy as np
from numpy import pi
from scipy.integrate import quad, nquad
from scipy.optimize import brentq, fsolve
from pathos.multiprocessing import ProcessingPool as Pool
from scipy.interpolate import interp1d, interp2d, RegularGridInterpolator
import pickle
from .interpolations.interps import *
from tqdm import tqdm

import matplotlib.pyplot as plt

from fastcore.test import *

# %% ../nbs/04_utils.ipynb 4
#Put 0 indent assignments so that variables will be in __all__
ut_interp = ut_interp
ut_interp_mw = ut_interp_mw
ut_interp_rho = ut_interp_rho
u_fwhm_interp = u_fwhm_interp
ut_interp_nfw = ut_interp_nfw
ut_interp_nfw_101 = ut_interp_nfw_101
ut_interp_axion = ut_interp_axion
ut_interp_dressed = ut_interp_dressed

# %% ../nbs/04_utils.ipynb 5
def dist_mw(d: float, # distance from the Sun in kpc
            ) -> float: #distance to the MW center in kpc
    return np.sqrt(d**2 + rEarth**2 - 2*d*rEarth*np.cos(np.radians(l))*np.cos(np.radians(b)))

def dist_m31(d: float, # distance from the Sun in kpc
             ) -> float: #distance to the M31 center in kpc
            return dsM31 - d

def density_mw(r: float, # distance to MW center in kpc
                ) -> float: # DM density in Msun/kpc^3
    return rhoc / ((r/rs) * (1 + r/rs)**2)

def density_m31(r: float, # distance to M31 center in kpc
                ) -> float: # DM density in Msun/kpc^3
    return rhocM31 / ((r/rsM31) * (1 + r/rsM31)**2)

def mass_enclosed_mw(r: float  # distance to MW center in kpc
                      ) -> float : # enclosed DM mass in Msun
    return 4*pi * rhoc * rs**3 * (np.log(1 + r/rs) - (r/rs)/(1 + r/rs))

def mass_enclosed_m31(r: float  # distance to M31 center in kpc
                        ) -> float : # enclosed DM mass in Msun
    return 4*pi * rhocM31 * rsM31**3 * (np.log(1 + r/rsM31) - (r/rsM31)/(1 + r/rsM31))

def velocity_dispersion_mw(r: float, # distance from the MW center in kpc
                        ) -> float: # velocity dispersion in km/s
    if r == 0:
        return 0
    return np.sqrt(G * mass_enclosed_mw(r) / r) 

def velocity_dispersion_m31(r: float, # distance from the M31 center in kpc
                        ) -> float: # velocity dispersion in km/s
    if r == 0:
        return 0
    return np.sqrt(G * mass_enclosed_m31(r) / r)

def dist(d: float, # distance from the Sun in kpc
         ds: float = ds, # distance to the source in kpc
         ) -> float: #weighted lensing distance in kpc
    if d > ds:
         raise ValueError("Distance of lens must be less than source distance but d = " + str(d) + " and ds = " + str(ds))
    return d * (1 - d/ds)

def einstein_rad(d: float, # distance from the Sun in kpc
                 mass: float, # mass of the lens in Msun
                 ds: float, # distance to the source in kpc
                 ) -> float: # Einstein radius in kpc
    return (4 * G * mass * dist(d, ds)/c**2)**(1/2)

def velocity_radial(d: float, # distance from the Sun in kpc
                    mass: float, # mass of the lens in Msun
                    umin: float, # minimum impact parameter
                    t: float, # crossing time in hours
                    ut: float, # threshold impact parameter
                    ds: float, # distance to the source in kpc
                    ) -> float: # radial velocity in km/s
    return 2*einstein_rad(d, mass, ds) * (ut**2 - umin**2)**(1/2) / t * kpctokm

def event_duration(d: float, # distance from the Sun in kpc
                   mass: float, # mass of the lens in Msun
                   umin: float, # minimum impact parameter
                   v_rel: float, # transverse velocity in km/s
                   ut: float, # threshold impact parameter
                   ds: float, # distance to the source in kpc
                    ) -> float: # event duration in seconds
    # return 2*einstein_rad(d, mass)* kpctokm * (ut**2 - umin**2)**(1/2) / v_rel
    rho = rho_func(mass, d, ds)
    return 2* (1 - ((umin + rho)/ut)**2)**(1/2) * einstein_rad(d, mass) * kpctokm * (ut + rho) / v_rel 
     

# from below 16 of https://iopscience.iop.org/article/10.3847/1538-4357/ac07a8/pdf*)
# alphabar = 27 Degrees xp-axis is aligned with the major axis
# of the Galactic bar,where \[Alpha]bar=27\[Degree] is applied as the bar angle.
# galactocentric coordniates x', y', z' as function of d, distance from Sun

def get_primed_coords(d: float, # distance from Sun in km
                      l: float = l, # galactic longitude in degrees
                      b: float = b, # galactic latitude in degrees
                      )-> tuple:
    """Get galactocentric coordinates x', y' given galactic latitude and longitude l, b, and distance d
    """
    # convert angles from degrees to radians
    l_rad = np.deg2rad(l)
    b_rad = np.deg2rad(b)
    alpha_rad = np.deg2rad(alphabar)

    # calculate unrotated Cartesian coordinates
    x_unrot = rEarth - d * np.cos(b_rad) * np.cos(l_rad)
    y_unrot = d * np.cos(b_rad) * np.sin(l_rad)

    # rotate the coordinates
    x_prime = x_unrot * np.cos(alpha_rad) - y_unrot * np.sin(alpha_rad)
    y_prime = x_unrot * np.sin(alpha_rad) + y_unrot * np.cos(alpha_rad)

    z_prime = d * np.sin(b_rad)

    return x_prime, y_prime, z_prime

def scientific_format(x, pos):
    """
    Formats a number in scientific notation in latex
    """
    a, b = '{:.1e}'.format(x).split('e')
    b = int(b)
    return r'${} \times 10^{{{}}}$'.format(a, b)

# %% ../nbs/04_utils.ipynb 8
# Add finite size calculation following https://arxiv.org/pdf/1905.06066.pdf

# Compute 'w' parameter given the mass of the primordial black hole and the wavelength
def w_func(m_pbh, lam):
    return 5.98 * (m_pbh / 1e-10) * (lam / 6210)**(-1)

# Compute 'rho' parameter given the mass of the primordial black hole and the lens distance
def rho_func(m_pbh, dl, ds):
    if dl >= ds:
        # raise ValueError("dl must be less than ds to prevent division by zero.")
        return np.inf
    if dl == 0:
        dl = 1e-2
    x = dl / ds
    # return 5.9 * (m_pbh / 1e-10)**(-1/2) * (x / (1-x))**(1/2) 
    return 5.9 * (m_pbh / 1e-10)**(-1/2) * (x / (1-x))**(1/2) * (ds/770)**(-1/2)


# Compute magnification given the impact parameter 'u'
def magnification(u):
        '''
        Compute the magnification given the impact parameter 'u' in the point-source-point-lens limit.
        '''
        if u == 0:
            return np.inf
        else:
            return (u**2 + 2) / (u * (u**2 + 4)**0.5)

# Compute magnification in the wave optics regime
def magnification_wave(w, u):
    # Note this is taking the maximum value of the wave optics magnification
    # In reality, need to evaluate the hypergeometric function
    return np.minimum(magnification(u), np.pi * w / (1 - np.exp(-np.pi * w)))

# Compute displacement given 'x', 'y', and 'u'
def displacement(x, y, u):
    return ((x - u)**2 + y**2)**0.5

# Compute integrand in polar coordinates
def integrand_polar_wave(r, theta, w, u):
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return magnification_wave(w, displacement(x, y, u)) * r

def integrand_polar(r, theta, u):
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return magnification(displacement(x, y, u)) * r

# Compute finite magnification
def magnification_finite_wave(m_pbh, lam, u, dl, ds):
    w = w_func(m_pbh, lam)
    rho = rho_func(m_pbh, dl, ds)
    integrand = lambda r, theta: integrand_polar_wave(r, theta, w, u)
    result, _ = nquad(integrand, [[0, rho], [0, 2 * pi]])
    return result / (pi * rho**2)

def magnification_finite(m_pbh, u, dl, ds):
    rho = rho_func(m_pbh, dl, ds)
    integrand = lambda r, theta: integrand_polar(r, theta, u)
    result, _ = nquad(integrand, [[0, rho], [0, 2 * pi]])
    return result / (pi * rho**2)

# Compute 'u' at threshold
def u_t_finite(m_pbh, dl, ds, A_thresh = 1.34):
    func = lambda u: magnification_finite(m_pbh, u, dl, ds) - A_thresh
    u_min = 0
    u_max = 10

    try:
        return brentq(func, u_min, u_max)
    except ValueError:
        return 0
    
def u_t_point(A_thresh = 1.34):
    func = lambda u: magnification(u) - A_thresh
    u_min = 0
    u_max = 10
    try:
        return brentq(func, u_min, u_max) 
    except ValueError:
        return 0
    
def u_t_finite_wave(m_pbh, lam, dl, ds, A_thresh = 1.34):
    func = lambda u: magnification_finite_wave(m_pbh, lam, u, dl, ds) - A_thresh
    u_min = 0
    u_max = 10

    try:
        return brentq(func, u_min, u_max)
    except ValueError:
        return 0

m_low_interp = 1e-15
m_high_interp = 1e0

def make_ut_interp(n_points=40, ds = 770, A_thresh=1.34):
    d_arr = np.linspace(0, ds, n_points)
    m_arr = np.logspace(np.log10(m_low_interp), np.log10(m_high_interp), n_points) #solar masses

    def calc_ut_arr(m):
    # Calculate ut_arr for the current m
        return np.array([u_t_finite(m, d, ds, A_thresh=A_thresh) for d in d_arr])
    
    with Pool() as p:
        ut_values = list(p.map(calc_ut_arr, m_arr))

    # Convert ut_values to a 2D array
    ut_values = np.array(ut_values)
    
    # Create the 2D interpolation table
    ut_interp = interp2d(d_arr, m_arr, ut_values)
    return ut_interp

# %% ../nbs/04_utils.ipynb 9
def u_t_nfw(rho, r90):
    if r90 > ut_interp_nfw.grid[1][-1]:
        return 0
    elif r90 < ut_interp_nfw.grid[1][0]:
        r90 = ut_interp_nfw.grid[1][0]
    if rho > ut_interp_nfw.grid[0][-1]:
        return 0
    elif rho < ut_interp_nfw.grid[0][0]:
        rho = ut_interp_nfw.grid[0][0]
    return ut_interp_nfw((rho, r90))

def u_t_nfw_101(rho, r90):
    if r90 > ut_interp_nfw_101.grid[1][-1]:
        return 0
    elif r90 < ut_interp_nfw_101.grid[1][0]:
        r90 = ut_interp_nfw_101.grid[1][0]
    if rho > ut_interp_nfw_101.grid[0][-1]:
        return 0
    elif rho < ut_interp_nfw_101.grid[0][0]:
        rho = ut_interp_nfw_101.grid[0][0]
    return ut_interp_nfw_101((rho, r90))

def u_t_axion(rho, r90):
    if r90 > ut_interp_axion.grid[1][-1]:
        return 0
    elif r90 < ut_interp_axion.grid[1][0]:
        r90 = ut_interp_axion.grid[1][0]
    if rho > ut_interp_axion.grid[0][-1]:
        return 0
    elif rho < ut_interp_axion.grid[0][0]:
        rho = ut_interp_axion.grid[0][0]
    return ut_interp_axion((rho, r90))

def u_t_axion_101(rho, r90):
    if r90 > ut_interp_axion_101.grid[1][-1]:
        return 0
    elif r90 < ut_interp_axion_101.grid[1][0]:
        r90 = ut_interp_axion_101.grid[1][0]
    if rho > ut_interp_axion_101.grid[0][-1]:
        return 0
    elif rho < ut_interp_axion_101.grid[0][0]:
        rho = ut_interp_axion_101.grid[0][0]
    return ut_interp_axion_101((rho, r90))

def u_t_dressed(rho, r90):
    if r90 > ut_interp_dressed.grid[1][-1]:
        return 0
    elif r90 < ut_interp_dressed.grid[1][0]:
        r90 = ut_interp_dressed.grid[1][0]
    if rho > ut_interp_dressed.grid[0][-1]:
        return 0
    elif rho < ut_interp_dressed.grid[0][0]:
        rho = ut_interp_dressed.grid[0][0]
    return ut_interp_dressed((rho, r90))

def u_t_dressed_101(rho, r90):
    if r90 > ut_interp_dressed_101.grid[1][-1]:
        return 0
    elif r90 < ut_interp_dressed_101.grid[1][0]:
        r90 = ut_interp_dressed_101.grid[1][0]
    if rho > ut_interp_dressed_101.grid[0][-1]:
        return 0
    elif rho < ut_interp_dressed_101.grid[0][0]:
        rho = ut_interp_dressed_101.grid[0][0]
    return ut_interp_dressed_101((rho, r90))

# %% ../nbs/04_utils.ipynb 10
def u_t_nfw_thresh(rho, r90):
    if r90 > ut_interp_nfw_thresh.grid[1][-1]:
        return 0
    elif r90 < ut_interp_nfw_thresh.grid[1][0]:
        r90 = ut_interp_nfw_thresh.grid[1][0]
    if rho > ut_interp_nfw_thresh.grid[0][-1]:
        return 0
    elif rho < ut_interp_nfw_thresh.grid[0][0]:
        rho = ut_interp_nfw_thresh.grid[0][0]
    return ut_interp_nfw_thresh((rho, r90))

def u_t_axion_thresh(rho, r90):
    if r90 > ut_interp_axion_thresh.grid[1][-1]:
        return 0
    elif r90 < ut_interp_axion_thresh.grid[1][0]:
        r90 = ut_interp_axion_thresh.grid[1][0]
    if rho > ut_interp_axion_thresh.grid[0][-1]:
        return 0
    elif rho < ut_interp_axion_thresh.grid[0][0]:
        rho = ut_interp_axion_thresh.grid[0][0]
    return ut_interp_axion_thresh((rho, r90))

def u_t_dressed_thresh(rho, r90):
    if r90 > ut_interp_dressed_thresh.grid[1][-1]:
        return 0
    elif r90 < ut_interp_dressed_thresh.grid[1][0]:
        r90 = ut_interp_dressed_thresh.grid[1][0]
    if rho > ut_interp_dressed_thresh.grid[0][-1]:
        return 0
    elif rho < ut_interp_dressed_thresh.grid[0][0]:
        rho = ut_interp_dressed_thresh.grid[0][0]
    return ut_interp_dressed_thresh((rho, r90))



# %% ../nbs/04_utils.ipynb 17
#Ratio of angular extent of source and lens in plane of lens
# rho == theta_s/theta_l
rho_min = 0.1
# rho_max = 4.5
rho_max = 14.2

rhos = np.linspace(rho_min, rho_max, 300)
# rhos = np.linspace(rho_min, rho_max, 40)


A_thresh_min = 1.01
A_thresh_max = 1.34
A_threshs = np.linspace(A_thresh_min, A_thresh_max, 40)


# %% ../nbs/04_utils.ipynb 19
def ut_func_new(rho, A_thresh):
    # rho = np.asarray(rho) 
    rho = np.atleast_1d(rho) # Ensure rho is at least 1-dimensional
    rho[np.where(rho <= rho_min)] = rho_min 
    rho[np.where(rho >= rho_max)] = rho_max 
    # return ut_interp_rho(rho)
    return ut_interp_rho(rho, A_thresh)

# %% ../nbs/04_utils.ipynb 34
def f_func(x):
    return (2 + x)/(x*(4+x))**(0.5) - 1
    
def t_fwhm_gondolo(t_e, umin):
    deltab = f_func(umin**2)
    bw2 = 2* f_func(deltab)
    w_b = 2*(bw2 - umin**2)**(1/2)
    return w_b * t_e

def uhm_point(umin):
    if umin==0:
        return np.inf
    return (2* f_func(f_func(umin**2)))**(1/2)

def t_e_from_fwhm(t_fwhm, umin, finite, rho=None):
    if not finite:
        return t_fwhm / t_fwhm_gondolo(1, umin)
    else:
        u_fwhm = u_fwhm_interp([rho, umin])[0]
        if u_fwhm<=umin:
            print('u_fwhm<=umin')
            return 1e-3
        return t_fwhm / (2*(u_fwhm**2 - umin**2)**(1/2))
